@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # gPodder - A media aggregator and podcast client
-# Copyright (c) 2005-2012 Thomas Perl and the gPodder Team
+# Copyright (c) 2005-2018 The gPodder Team
 #
 # gPodder is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -38,19 +38,22 @@ logger = logging.getLogger(__name__)
 from gpodder.gtkui import draw
 
 import os
-import gtk
-import gobject
+from gi.repository import Gtk
+from gi.repository import GObject
+from gi.repository import GdkPixbuf
 import cgi
 import re
+import time
 
 try:
-    import gio
+    from gi.repository import Gio
     have_gio = True
 except ImportError:
     have_gio = False
 
 # ----------------------------------------------------------
 
+
 class GEpisode(model.PodcastEpisode):
     __slots__ = ()
 
@@ -91,29 +94,70 @@ class GEpisode(model.PodcastEpisode):
                 cgi.escape(downloaded_string), \
                 cgi.escape(self.channel.title))
 
+
 class GPodcast(model.PodcastChannel):
     __slots__ = ()
 
     EpisodeClass = GEpisode
 
+
 class Model(model.Model):
     PodcastClass = GPodcast
 
 # ----------------------------------------------------------
 
+
 # Singleton indicator if a row is a section
 class SeparatorMarker(object): pass
+
+
 class SectionMarker(object): pass
 
-class EpisodeListModel(gtk.ListStore):
+
+class BackgroundUpdate(object):
+    def __init__(self, model, episodes, include_description):
+        self.model = model
+        self.episodes = episodes
+        self.include_description = include_description
+        self.index = 0
+
+    def update(self):
+        model = self.model
+        include_description = self.include_description
+
+        started = time.time()
+        while self.episodes:
+            episode = self.episodes.pop(0)
+            base_fields = (
+                (model.C_URL, episode.url),
+                (model.C_TITLE, episode.title),
+                (model.C_EPISODE, episode),
+                (model.C_PUBLISHED_TEXT, episode.cute_pubdate()),
+                (model.C_PUBLISHED, episode.published),
+            )
+            update_fields = model.get_update_fields(episode, include_description)
+            model.set(model.get_iter((self.index,)), *(x for fields in (base_fields, update_fields)
+                                                       for pair in fields for x in pair))
+            self.index += 1
+
+            # Check for the time limit of 20 ms after each 50 rows processed
+            if self.index % 50 == 0 and (time.time() - started) > 0.02:
+                break
+
+        return bool(self.episodes)
+
+
+class EpisodeListModel(Gtk.ListStore):
     C_URL, C_TITLE, C_FILESIZE_TEXT, C_EPISODE, C_STATUS_ICON, \
             C_PUBLISHED_TEXT, C_DESCRIPTION, C_TOOLTIP, \
             C_VIEW_SHOW_UNDELETED, C_VIEW_SHOW_DOWNLOADED, \
             C_VIEW_SHOW_UNPLAYED, C_FILESIZE, C_PUBLISHED, \
             C_TIME, C_TIME_VISIBLE, C_TOTAL_TIME, \
-            C_LOCKED = range(17)
+            C_LOCKED = list(range(17))
 
-    VIEW_ALL, VIEW_UNDELETED, VIEW_DOWNLOADED, VIEW_UNPLAYED = range(4)
+    VIEW_ALL, VIEW_UNDELETED, VIEW_DOWNLOADED, VIEW_UNPLAYED = list(range(4))
+
+    VIEWS = ['VIEW_ALL', 'VIEW_UNDELETED', 'VIEW_DOWNLOADED', 'VIEW_UNPLAYED']
 
     # In which steps the UI is updated for "loading" animations
     _UI_UPDATE_STEP = .03
@@ -121,10 +165,12 @@ class EpisodeListModel(gtk.ListStore):
     # Steps for the "downloading" icon progress
     PROGRESS_STEPS = 20
 
-    def __init__(self, on_filter_changed=lambda has_episodes: None):
-        gtk.ListStore.__init__(self, str, str, str, object, \
+    def __init__(self, config, on_filter_changed=lambda has_episodes: None):
+        Gtk.ListStore.__init__(self, str, str, str, object, \
                 str, str, str, str, bool, bool, bool, \
-                gobject.TYPE_INT64, int, str, bool, int, bool)
+                GObject.TYPE_INT64, GObject.TYPE_INT64, str, bool, GObject.TYPE_INT64, bool)
+
+        self._config = config
 
         # Callback for when the filter / list changes, gets one parameter
         # (has_episodes) that is True if the list has any episodes
@@ -132,7 +178,7 @@ class EpisodeListModel(gtk.ListStore):
 
         # Filter to allow hiding some episodes
         self._filter = self.filter_new()
-        self._sorter = gtk.TreeModelSort(self._filter)
+        self._sorter = Gtk.TreeModelSort(self._filter)
         self._view_mode = self.VIEW_ALL
         self._search_term = None
         self._search_term_eql = None
@@ -145,8 +191,11 @@ class EpisodeListModel(gtk.ListStore):
         self.ICON_VIDEO_FILE = 'video-x-generic'
         self.ICON_IMAGE_FILE = 'image-x-generic'
         self.ICON_GENERIC_FILE = 'text-x-generic'
-        self.ICON_DOWNLOADING = gtk.STOCK_GO_DOWN
-        self.ICON_DELETED = gtk.STOCK_DELETE
+        self.ICON_DOWNLOADING = Gtk.STOCK_GO_DOWN
+        self.ICON_DELETED = Gtk.STOCK_DELETE
+
+        self.background_update = None
+        self.background_update_tag = None
 
         if 'KDE_FULL_SESSION' in os.environ:
             # Workaround until KDE adds all the freedesktop icons
@@ -154,14 +203,13 @@ class EpisodeListModel(gtk.ListStore):
             #     http://gpodder.org/bug/553
             self.ICON_DELETED = 'archive-remove'
 
-
     def _format_filesize(self, episode):
         if episode.file_size > 0:
             return util.format_filesize(episode.file_size, digits=1)
         else:
             return None
 
-    def _filter_visible_func(self, model, iter):
+    def _filter_visible_func(self, model, iter, misc):
         # If searching is active, set visibility based on search text
         if self._search_term is not None:
             episode = model.get_value(iter, self.C_EPISODE)
@@ -170,7 +218,7 @@ class EpisodeListModel(gtk.ListStore):
 
             try:
                 return self._search_term_eql.match(episode)
-            except Exception, e:
+            except Exception as e:
                 return True
 
         if self._view_mode == self.VIEW_ALL:
@@ -227,23 +275,25 @@ class EpisodeListModel(gtk.ListStore):
 
     def _format_description(self, episode, include_description=False):
         title = episode.trimmed_title
-        a, b = '', ''
-        if episode.state != gpodder.STATE_DELETED and episode.is_new:
-            a, b = '<b>', '</b>'
-        if include_description and self._all_episodes_view:
-            return '%s%s%s\n%s' % (a, cgi.escape(title), b,
-                    _('from %s') % cgi.escape(episode.channel.title))
-        elif include_description:
-            description = episode.one_line_description()
-            if description.startswith(title):
-                description = description[len(title):].strip()
-            return '%s%s%s\n%s' % (a, cgi.escape(title), b,
-                    cgi.escape(description))
-        else:
-            return ''.join((a, cgi.escape(title), b))
 
-    def replace_from_channel(self, channel, include_description=False,
-            treeview=None):
+        if episode.state != gpodder.STATE_DELETED and episode.is_new:
+            yield '<b>'
+            yield cgi.escape(title)
+            yield '</b>'
+        else:
+            yield cgi.escape(title)
+
+        if include_description:
+            yield '\n'
+            if self._all_episodes_view:
+                yield _('from %s') % cgi.escape(episode.channel.title)
+            else:
+                description = episode.one_line_description()
+                if description.startswith(title):
+                    description = description[len(title):].strip()
+                yield cgi.escape(description)
+
+    def replace_from_channel(self, channel, include_description=False):
         """
         Add episode from the given channel to this model.
         Downloading should be a callback.
@@ -254,9 +304,6 @@ class EpisodeListModel(gtk.ListStore):
         # Remove old episodes in the list store
         self.clear()
 
-        if treeview is not None:
-            util.idle_add(treeview.queue_draw)
-
         self._all_episodes_view = getattr(channel, 'ALL_EPISODES_PROXY', False)
 
         # Avoid gPodder bug 1291
@@ -265,36 +312,42 @@ class EpisodeListModel(gtk.ListStore):
         else:
             episodes = channel.get_all_episodes()
 
-        if not isinstance(episodes, list):
-            episodes = list(episodes)
-        count = len(episodes)
+        # Always make a copy, so we can pass the episode list to BackgroundUpdate
+        episodes = list(episodes)
 
-        for position, episode in enumerate(episodes):
-            iter = self.append((episode.url, \
-                    episode.title, \
-                    self._format_filesize(episode), \
-                    episode, \
-                    None, \
-                    episode.cute_pubdate(), \
-                    '', \
-                    '', \
-                    True, \
-                    True, \
-                    True, \
-                    episode.file_size, \
-                    episode.published, \
-                    episode.get_play_info_string(), \
-                    bool(episode.total_time), \
-                    episode.total_time, \
-                    episode.archive))
+        for _ in range(len(episodes)):
+            self.append()
 
-            self.update_by_iter(iter, include_description)
+        self._update_from_episodes(episodes, include_description)
 
-        self._on_filter_changed(self.has_episodes())
+    def _update_from_episodes(self, episodes, include_description):
+        if self.background_update_tag is not None:
+            GObject.source_remove(self.background_update_tag)
+
+        self.background_update = BackgroundUpdate(self, episodes, include_description)
+        self.background_update_tag = GObject.idle_add(self._update_background)
+
+    def _update_background(self):
+        if self.background_update is not None:
+            if self.background_update.update():
+                return True
+
+            self.background_update = None
+            self.background_update_tag = None
+            self._on_filter_changed(self.has_episodes())
+
+        return False
 
     def update_all(self, include_description=False):
-        for row in self:
-            self.update_by_iter(row.iter, include_description)
+        if self.background_update is None:
+            episodes = [row[self.C_EPISODE] for row in self]
+        else:
+            # Update all episodes that have already been initialized...
+            episodes = [row[self.C_EPISODE] for index, row in enumerate(self) if index < self.background_update.index]
+            # ...and also include episodes that still need to be initialized
+            episodes.extend(self.background_update.episodes)
+
+        self._update_from_episodes(episodes, include_description)
 
     def update_by_urls(self, urls, include_description=False):
         for row in self:
@@ -304,13 +357,11 @@ class EpisodeListModel(gtk.ListStore):
     def update_by_filter_iter(self, iter, include_description=False):
         # Convenience function for use by "outside" methods that use iters
         # from the filtered episode list model (i.e. all UI things normally)
-        iter = self._sorter.convert_iter_to_child_iter(None, iter)
+        iter = self._sorter.convert_iter_to_child_iter(iter)
         self.update_by_iter(self._filter.convert_iter_to_child_iter(iter),
                 include_description)
 
-    def update_by_iter(self, iter, include_description=False):
-        episode = self.get_value(iter, self.C_EPISODE)
-
+    def get_update_fields(self, episode, include_description):
         show_bullet = False
         show_padlock = False
         show_missing = False
@@ -319,7 +370,7 @@ class EpisodeListModel(gtk.ListStore):
         view_show_undeleted = True
         view_show_downloaded = False
         view_show_unplayed = False
-        icon_theme = gtk.icon_theme_get_default()
+        icon_theme = Gtk.IconTheme.get_default()
 
         if episode.downloading:
             tooltip.append('%s %d%%' % (_('Downloading'),
@@ -365,9 +416,9 @@ class EpisodeListModel(gtk.ListStore):
 
                 # Try to find a themed icon for this file
                 if filename is not None and have_gio:
-                    file = gio.File(filename)
+                    file = Gio.File.new_for_path(filename)
                     if file.query_exists():
-                        file_info = file.query_info('*')
+                        file_info = file.query_info('*', Gio.FileQueryInfoFlags.NONE, None)
                         icon = file_info.get_icon()
                         for icon_name in icon.get_names():
                             if icon_theme.has_icon(icon_name):
@@ -404,20 +455,26 @@ class EpisodeListModel(gtk.ListStore):
 
         tooltip = ', '.join(tooltip)
 
-        description = self._format_description(episode, include_description)
-        self.set(iter, \
-                self.C_STATUS_ICON, status_icon, \
-                self.C_VIEW_SHOW_UNDELETED, view_show_undeleted, \
-                self.C_VIEW_SHOW_DOWNLOADED, view_show_downloaded, \
-                self.C_VIEW_SHOW_UNPLAYED, view_show_unplayed, \
-                self.C_DESCRIPTION, description, \
-                self.C_TOOLTIP, tooltip, \
-                self.C_TIME, episode.get_play_info_string(duration_only=True), \
-                self.C_TIME_VISIBLE, bool(episode.total_time), \
-                self.C_TOTAL_TIME, episode.total_time, \
-                self.C_LOCKED, episode.archive, \
-                self.C_FILESIZE_TEXT, self._format_filesize(episode), \
-                self.C_FILESIZE, episode.file_size)
+        description = ''.join(self._format_description(episode, include_description))
+        return (
+                (self.C_STATUS_ICON, status_icon),
+                (self.C_VIEW_SHOW_UNDELETED, view_show_undeleted),
+                (self.C_VIEW_SHOW_DOWNLOADED, view_show_downloaded),
+                (self.C_VIEW_SHOW_UNPLAYED, view_show_unplayed),
+                (self.C_DESCRIPTION, description),
+                (self.C_TOOLTIP, tooltip),
+                (self.C_TIME, episode.get_play_info_string()),
+                (self.C_TIME_VISIBLE, bool(episode.total_time)),
+                (self.C_TOTAL_TIME, episode.total_time),
+                (self.C_LOCKED, episode.archive),
+                (self.C_FILESIZE_TEXT, self._format_filesize(episode)),
+                (self.C_FILESIZE, episode.file_size),
+        )
+
+    def update_by_iter(self, iter, include_description=False):
+        episode = self.get_value(iter, self.C_EPISODE)
+        if episode is not None:
+            self.set(iter, *(x for pair in self.get_update_fields(episode, include_description) for x in pair))
 
 
 class PodcastChannelProxy(object):
@@ -427,7 +484,7 @@ class PodcastChannelProxy(object):
         self._db = db
         self._config = config
         self.channels = channels
-        self.title =  _('All episodes')
+        self.title = _('All episodes')
         self.description = _('from all podcasts')
         #self.parse_error = ''
         self.url = ''
@@ -438,6 +495,8 @@ class PodcastChannelProxy(object):
         self.auth_username = None
         self.auth_password = None
         self.pause_subscription = False
+        self.sync_to_mp3_player = False
+        self.cover_thumb = None
         self.auto_archive_episodes = False
 
     def get_statistics(self):
@@ -449,13 +508,16 @@ class PodcastChannelProxy(object):
         return Model.sort_episodes_by_pubdate((e for c in self.channels
                 for e in c.get_all_episodes()), True)
 
+    def save(self):
+        pass
 
-class PodcastListModel(gtk.ListStore):
+
+class PodcastListModel(Gtk.ListStore):
     C_URL, C_TITLE, C_DESCRIPTION, C_PILL, C_CHANNEL, \
             C_COVER, C_ERROR, C_PILL_VISIBLE, \
             C_VIEW_SHOW_UNDELETED, C_VIEW_SHOW_DOWNLOADED, \
             C_VIEW_SHOW_UNPLAYED, C_HAS_EPISODES, C_SEPARATOR, \
-            C_DOWNLOADS, C_COVER_VISIBLE, C_SECTION = range(16)
+            C_DOWNLOADS, C_COVER_VISIBLE, C_SECTION = list(range(16))
 
     SEARCH_COLUMNS = (C_TITLE, C_DESCRIPTION, C_SECTION)
 
@@ -464,8 +526,8 @@ class PodcastListModel(gtk.ListStore):
         return model.get_value(iter, cls.C_SEPARATOR)
 
     def __init__(self, cover_downloader):
-        gtk.ListStore.__init__(self, str, str, str, gtk.gdk.Pixbuf, \
-                object, gtk.gdk.Pixbuf, str, bool, bool, bool, bool, \
+        Gtk.ListStore.__init__(self, str, str, str, GdkPixbuf.Pixbuf, \
+                object, GdkPixbuf.Pixbuf, str, bool, bool, bool, bool, \
                 bool, bool, int, bool, str)
 
         # Filter to allow hiding some episodes
@@ -480,7 +542,7 @@ class PodcastListModel(gtk.ListStore):
 
         self.ICON_DISABLED = 'gtk-media-pause'
 
-    def _filter_visible_func(self, model, iter):
+    def _filter_visible_func(self, model, iter, misc):
         # If searching is active, set visibility based on search text
         if self._search_term is not None:
             if model.get_value(iter, self.C_CHANNEL) == SectionMarker:
@@ -554,14 +616,14 @@ class PodcastListModel(gtk.ListStore):
         if pixbuf.get_width() > self._max_image_side:
             f = float(self._max_image_side)/pixbuf.get_width()
             (width, height) = (int(pixbuf.get_width()*f), int(pixbuf.get_height()*f))
-            pixbuf = pixbuf.scale_simple(width, height, gtk.gdk.INTERP_BILINEAR)
+            pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
             changed = True
 
         # Resize if too high
         if pixbuf.get_height() > self._max_image_side:
             f = float(self._max_image_side)/pixbuf.get_height()
             (width, height) = (int(pixbuf.get_width()*f), int(pixbuf.get_height()*f))
-            pixbuf = pixbuf.scale_simple(width, height, gtk.gdk.INTERP_BILINEAR)
+            pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
             changed = True
 
         if changed:
@@ -578,7 +640,7 @@ class PodcastListModel(gtk.ListStore):
 
     def _overlay_pixbuf(self, pixbuf, icon):
         try:
-            icon_theme = gtk.icon_theme_get_default()
+            icon_theme = Gtk.IconTheme.get_default()
             emblem = icon_theme.load_icon(icon, self._max_image_side/2, 0)
             (width, height) = (emblem.get_width(), emblem.get_height())
             xpos = pixbuf.get_width() - width
@@ -589,18 +651,48 @@ class PodcastListModel(gtk.ListStore):
                 (width, height) = (emblem.get_width(), emblem.get_height())
                 xpos = pixbuf.get_width() - width
                 ypos = pixbuf.get_height() - height
-            emblem.composite(pixbuf, xpos, ypos, width, height, xpos, ypos, 1, 1, gtk.gdk.INTERP_BILINEAR, 255)
+            emblem.composite(pixbuf, xpos, ypos, width, height, xpos, ypos, 1, 1, GdkPixbuf.InterpType.BILINEAR, 255)
         except:
             pass
 
         return pixbuf
 
+    def _get_cached_thumb(self, channel):
+        if channel.cover_thumb is None:
+            return None
+
+        try:
+            loader = GdkPixbuf.PixbufLoader()
+            loader.write(channel.cover_thumb)
+            loader.close()
+            return loader.get_pixbuf()
+        except Exception as e:
+            logger.warn('Could not load cached cover art for %s', channel.url, exc_info=True)
+            channel.cover_thumb = None
+            channel.save()
+            return None
+
+    def _save_cached_thumb(self, channel, pixbuf):
+        bufs = []
+
+        def save_callback(buf, length, user_data):
+            user_data.append(buf)
+            return True
+        pixbuf.save_to_callbackv(save_callback, bufs, 'png', [None], [])
+        channel.cover_thumb = bytes(b''.join(bufs))
+        channel.save()
+
     def _get_cover_image(self, channel, add_overlay=False):
         if self._cover_downloader is None:
             return None
 
-        pixbuf = self._cover_downloader.get_cover(channel, avoid_downloading=True)
-        pixbuf_overlay = self._resize_pixbuf(channel.url, pixbuf)
+        pixbuf_overlay = self._get_cached_thumb(channel)
+
+        if pixbuf_overlay is None:
+            pixbuf = self._cover_downloader.get_cover(channel, avoid_downloading=True)
+            pixbuf_overlay = self._resize_pixbuf(channel.url, pixbuf)
+            self._save_cached_thumb(channel, pixbuf_overlay)
+
         if add_overlay and channel.pause_subscription:
             pixbuf_overlay = self._overlay_pixbuf(pixbuf_overlay, self.ICON_DISABLED)
             pixbuf_overlay.saturate_and_pixelate(pixbuf_overlay, 0.0, False)
@@ -719,7 +811,7 @@ class PodcastListModel(gtk.ListStore):
     def iter_is_first_row(self, iter):
         iter = self._filter.convert_iter_to_child_iter(iter)
         path = self.get_path(iter)
-        return (path == (0,))
+        return (path == Gtk.TreePath.new_first())
 
     def update_by_filter_iter(self, iter):
         self.update_by_iter(self._filter.convert_iter_to_child_iter(iter))
@@ -748,8 +840,11 @@ class PodcastListModel(gtk.ListStore):
                     if isinstance(c, GPodcast) and c.section == section]
 
             # Calculate the stats over all podcasts of this section
-            total, deleted, new, downloaded, unplayed = map(sum,
-                    zip(*[c.get_statistics() for c in channels]))
+            if len(channels) is 0:
+                total = deleted = new = downloaded = unplayed = 0
+            else:
+                total, deleted, new, downloaded, unplayed = list(map(sum,
+                        list(zip(*[c.get_statistics() for c in channels]))))
 
             # We could customized the section header here with the list
             # of channels and their stats (i.e. add some "new" indicator)
@@ -764,7 +859,7 @@ class PodcastListModel(gtk.ListStore):
                 self.C_VIEW_SHOW_UNPLAYED, unplayed + new > 0)
 
         if (not isinstance(channel, GPodcast) and
-            not isinstance(channel, PodcastChannelProxy)):
+                not isinstance(channel, PodcastChannelProxy)):
             return
 
         total, deleted, new, downloaded, unplayed = channel.get_statistics()
@@ -779,7 +874,7 @@ class PodcastListModel(gtk.ListStore):
                 self.C_SECTION, channel.section, \
                 self.C_ERROR, self._format_error(channel), \
                 self.C_PILL, pill_image, \
-                self.C_PILL_VISIBLE, pill_image != None, \
+                self.C_PILL_VISIBLE, pill_image is not None, \
                 self.C_VIEW_SHOW_UNDELETED, total - deleted > 0, \
                 self.C_VIEW_SHOW_DOWNLOADED, downloaded + new > 0, \
                 self.C_VIEW_SHOW_UNPLAYED, unplayed + new > 0, \
@@ -792,8 +887,15 @@ class PodcastListModel(gtk.ListStore):
             del self._cover_cache[podcast_url]
 
     def add_cover_by_channel(self, channel, pixbuf):
+        if pixbuf is None:
+            return
+        # Remove older images from cache
+        self.clear_cover_cache(channel.url)
+
         # Resize and add the new cover image
         pixbuf = self._resize_pixbuf(channel.url, pixbuf)
+        self._save_cached_thumb(channel, pixbuf)
+
         if channel.pause_subscription:
             pixbuf = self._overlay_pixbuf(pixbuf, self.ICON_DISABLED)
             pixbuf.saturate_and_pixelate(pixbuf, 0.0, False)
